#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from typing import (
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Union,
)

from .indexer import CompressedIndexedText


class FunnyPassphraseGenerator(object):
    """
    This class takes as input a sequence of CompressedIndexedText
    instances, which can be considered groups of words.
    It generates a passphrase in the form of a random word from each
    group of words. If the number of words is larger than the number of
    groups, the generator loops over the groups of words.
    """

    def __init__(self, *args: CompressedIndexedText, **kwargs: CompressedIndexedText):
        cit: MutableSequence[CompressedIndexedText]
        cit_dict: MutableMapping[str, CompressedIndexedText]
        if args:
            cit = list(arg for arg in args)
            cit_dict = dict((str(i_arg), arg) for i_arg, arg in enumerate(args))
        else:
            cit = list()
            cit_dict = dict()

        if kwargs:
            cit_dict.update(kwargs)
            cit.extend(kwargs.values())

        self.cit = cit
        self.cit_dict: MutableMapping[str, CompressedIndexedText] = cit_dict

    def __len__(self) -> int:
        """
        It returns the number of different word sets
        """
        return len(self.cit)

    def word_set_tags(self) -> Sequence[Union[str, int]]:
        """
        It returns the symbolic names of the word sets
        """
        return list(self.cit_dict.keys())

    def generate(
        self, num: int = 5, subset: Optional[Sequence[Union[str, int]]] = None
    ) -> Iterator[str]:
        """
        This method allows generating a passphrase, based either in all
        the word sets received on instantiation, or using an ordered subset
        """
        cit_dict: Union[
            Sequence[CompressedIndexedText],
            Mapping[str, CompressedIndexedText],
        ]
        subset_l: Union[Sequence[int], Sequence[str]]
        if subset:
            subset_l = list(str(s) for s in subset)
            cit_dict = self.cit_dict
            for s_name in subset_l:
                if s_name not in cit_dict:
                    raise KeyError(
                        f"Word set {s_name} is not among the available ones ({cit_dict.keys()})"
                    )

            # Duplicated block to avoid mypy complains
            for cit_idx in range(num):
                real_cit_idx = cit_idx % len(subset_l)
                cI = cit_dict[subset_l[real_cit_idx]]
                yield cI.get_random_line()
        else:
            # Suboptimal but effective
            subset_l = list(range(len(self)))
            cit = self.cit

            # Duplicated block to avoid mypy complains
            for cit_idx in range(num):
                real_cit_idx = cit_idx % len(subset_l)
                cI = cit[subset_l[real_cit_idx]]
                yield cI.get_random_line()

    def generate_passphrase(
        self,
        sep: str = " ",
        num: int = 5,
        subset: Optional[Sequence[Union[str, int]]] = None,
    ) -> str:
        """
        This method allows generating a passphrase,
        using the given separator
        """
        return sep.join(self.generate(num=num, subset=subset))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import io
import random
import shutil
import tempfile

from typing import (
    cast,
    Optional,
    Sequence,
    TextIO,
    Union,
)

from typing_extensions import Final
import xz  # type: ignore


class CompressedIndexedText(object):
    """
    Compress textual files, generating also a compressed index
    of the starting lines Index textual files
    """

    INT_BIN_SIZE: Final[int] = 8

    def __init__(self, cfiles: Sequence[str], encoding: str = "utf-8"):
        self.encoding = encoding
        self.cf_l = []
        self._max_lines_l = []
        num_lines = 0
        for cfile in cfiles:
            cF = xz.open(cfile, mode="rb")

            # Now, get the number of members
            if len(cF.stream_boundaries) != 2:
                raise ValueError(f"Unexpected number of members in {cfile}")

            f_length = cF.seek(0, io.SEEK_END)
            if (f_length - cF.stream_boundaries[1]) % self.INT_BIN_SIZE:
                raise ValueError(f"Unexpected mismatch in elements from {cfile}")
            _num_lines = (f_length - cF.stream_boundaries[1]) // self.INT_BIN_SIZE

            self.cf_l.append(cF)
            num_lines += _num_lines
            # Limits in number of lines
            self._max_lines_l.append(num_lines)

        self._num_lines = num_lines

    def get_line(self, lineno: int) -> str:
        if lineno < 0 or lineno >= self._num_lines:
            raise ValueError(
                f"Requested line {lineno} is out of range [0, {self._num_lines})"
            )

        cF = None
        prev_max = 0
        lineno_effective = lineno
        for the_cF, linemax in zip(self.cf_l, self._max_lines_l):
            if lineno < linemax:
                cF = the_cF
                lineno_effective = lineno - prev_max
                break
            prev_max = linemax

        if cF is None:
            raise ValueError(f"Corrupted code?")

        cF.seek(
            cF.stream_boundaries[1] + lineno_effective * self.INT_BIN_SIZE, io.SEEK_SET
        )
        boffset = cF.read(self.INT_BIN_SIZE)
        offset = int.from_bytes(boffset, byteorder="big", signed=False)
        if offset >= cF.stream_boundaries[1]:
            raise ValueError(f"Corrupted index")

        # Seek the pos
        cF.seek(offset, io.SEEK_SET)

        # And return the line, translated to the encoding
        line: str = cF.readline().decode(self.encoding).rstrip()
        return line

    @property
    def num_lines(self) -> int:
        return self._num_lines

    def get_random_line(self) -> str:
        rand_line = random.randint(0, self._num_lines - 1)
        return self.get_line(rand_line)

    @classmethod
    def IndexTextStream(
        cls,
        instream: Union[io.IOBase, TextIO],
        outfile: str,
        encoding: str = "utf-8",
    ) -> "CompressedIndexedText":
        """
        It is assumed instream is a file-like object
        opened in binary mode
        or something which implements readline method
        returning bytes and tell method
        in order to read line by line
        """

        trans_to_bytes = isinstance(instream, io.TextIOBase)

        with xz.open(outfile, mode="wb") as output, tempfile.SpooledTemporaryFile(
            max_size=1024 * 1024
        ) as output_idx:
            # posarr = []
            pos = instream.tell()
            line = instream.readline()
            bline: bytes
            if trans_to_bytes:
                assert isinstance(line, str)
                bline = line.encode(encoding)
            else:
                assert isinstance(line, bytes)
                bline = line
            # First, write the lines taking note of the offsets
            lastb = b""
            while len(bline) > 0:
                lastb = bline[-1:]
                output.write(bline)
                # posarr.append(pos)
                output_idx.write(
                    (pos).to_bytes(cls.INT_BIN_SIZE, byteorder="big", signed=False)
                )
                pos = instream.tell()
                line = instream.readline()
                if trans_to_bytes:
                    assert isinstance(line, str)
                    bline = line.encode(encoding)
                else:
                    assert isinstance(line, bytes)
                    bline = line

            # This is needed to avoid corner cases
            if lastb != b"\n":
                output.write(b"\n")

            # Now, write the index in a separate stream
            output.change_stream()
            output_idx.seek(0, io.SEEK_SET)
            shutil.copyfileobj(output_idx, output)
            # for pos in posarr:
            #    output.write((pos).to_bytes(cls.INT_BIN_SIZE, byteorder='big', signed=False))

        return cls(cfiles=[outfile], encoding=encoding)

    @classmethod
    def IndexTextFile(
        cls, infile: str, outfile: str, encoding: str = "utf-8"
    ) -> "CompressedIndexedText":
        with open(infile, mode="rb") as inpF:
            return cls.IndexTextStream(inpF, outfile, encoding=encoding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

__author__ = "José M. Fernández <https://orcid.org/0000-0002-4806-5140>"
__copyright__ = "© 2022 Barcelona Supercomputing Center (BSC), ES"
__license__ = "LGPL-2.1"

# https://www.python.org/dev/peps/pep-0396/
__version__ = "0.2.1"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import sys
from . import CompressedIndexedText


def main() -> None:
    ap = argparse.ArgumentParser(
        description="Funny passphrase indexer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # ap.add_argument('-e','--encoding', dest="encoding", help="Which encoding is expected in ")
    ap.add_argument(
        "input", help="Input file. If the name is '-', standard input will be used"
    )
    ap.add_argument("output", help="Output file")
    args = ap.parse_args()

    print(
        f"Compressing, indexing and writing all the text contents from {args.input} to {args.output}",
        file=sys.stderr,
    )

    if args.input == "-":
        cit = CompressedIndexedText.IndexTextStream(sys.stdin, args.output)
    else:
        cit = CompressedIndexedText.IndexTextFile(args.input, args.output)

    print(f"Number of indexed lines in {args.output}: {cit.num_lines}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import re
import os
import sys
import setuptools  # type: ignore

# In this way, we are sure we are getting
# the installer's version of the library
# not the system's one
setupDir = os.path.dirname(__file__)
sys.path.insert(0, setupDir)

from funny_passphrase import __version__ as funny_passphrase_version
from funny_passphrase import __author__ as funny_passphrase_author
from funny_passphrase import __license__ as funny_passphrase_license

# Populating the long description
readme_path = os.path.join(setupDir, "README.md")
with open(readme_path, "r") as fh:
    long_description = fh.read()

# Populating the install requirements
requirements = []
requirements_path = os.path.join(setupDir, "requirements.txt")
if os.path.exists(requirements_path):
    with open(requirements_path) as f:
        egg = re.compile(r"#[^#]*egg=([^=&]+)")
        for line in f.read().splitlines():
            m = egg.search(line)
            requirements.append(line if m is None else m.group(1))

setuptools.setup(
    name="funny-passphrase",
    version=funny_passphrase_version,
    author=funny_passphrase_author,
    author_email="jose.m.fernandez@bsc.es",
    license=funny_passphrase_license,
    description="A library to index text files, and generate passphrases from them",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/inab/funny-passphrase",
    project_urls={"Bug Tracker": "https://github.com/inab/funny-passphrase/issues"},
    packages=setuptools.find_packages(),
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "fp-indexer=funny_passphrase.indexer.__main__:main",
            "funny-passphrase=funny_passphrase.__main__:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# funny_passphrase, a library to index text files, and generate
# passphrases from them.
# Copyright (C) 2022 Barcelona Supercomputing Center, José M. Fernández
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import sys
from .generator import FunnyPassphraseGenerator
from .indexer import CompressedIndexedText


def main() -> None:
    ap = argparse.ArgumentParser(
        description="Funny passphrase generator",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # ap.add_argument('-e','--encoding', dest="encoding", help="Which encoding is expected in ")
    ap.add_argument(
        "--wg",
        dest="wgs",
        action="append",
        required=True,
        nargs="+",
        type=str,
        help="The indexed files which integrate this group",
    )
    ap.add_argument(
        "-w",
        "--words",
        action="store",
        required=True,
        type=int,
        help="Number of words to generate in each passphrase",
    )
    ap.add_argument(
        "-n",
        "--num",
        action="store",
        default=1,
        type=int,
        help="Number of passphrases to generate",
    )
    ap.add_argument(
        "-o",
        "--output",
        action="store",
        default="-",
        type=str,
        help="Output file (default stdout)",
    )
    args = ap.parse_args()

    ws_list = []
    if args.wgs is None:
        sys.exit(1)
    for i_ws, ws in enumerate(args.wgs):
        print(f"* Group {i_ws} with {len(ws)} word sets", file=sys.stderr)
        ws_list.append(CompressedIndexedText(ws))

    fpg = FunnyPassphraseGenerator(*ws_list)
    print(f"Writing the {args.num} passphrases to {args.output}", file=sys.stderr)
    if args.output == "-":
        out = sys.stdout
    else:
        out = open(args.output, mode="w", encoding="utf-8")
    try:
        for _ in range(args.num):
            print(fpg.generate_passphrase(sep=" ", num=args.words), file=out)
    finally:
        if out != sys.stdout:
            out.close()


if __name__ == "__main__":
    main()
